// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <string.h>
#include <unistd.h>

#include "source_delay.h"
#include "log.h"
#include "utils.h"
#include "picio.h"
#include "filter.h"

source_delay::source_delay(const std::string & id, const std::string & descr, const std::string & exec_failure, source *const s, const int jpeg_quality, const int n_frames, const double max_fps, resize *const r, const int resize_w, const int resize_h, const int ll, const double timeout, std::vector<filter *> *const filters) : source(id, descr, exec_failure, max_fps, r, resize_w, resize_h, ll, timeout, filters), s(s), jpeg_quality(jpeg_quality), n_frames(n_frames), max_fps(max_fps)
{
}

source_delay::~source_delay()
{
	for(h_data_t hd : frames)
		free(hd.data);

	free(prev_frame);
}

void source_delay::operator()()
{
	log(id, LL_INFO, "source_delay thread started");

	set_thread_name("source_delay");

	const uint64_t interval = max_fps > 0.0 ? 1.0 / max_fps * 1000.0 * 1000.0 : 0;

	s -> register_user();

	for(;!local_stop_flag;)
	{
		uint64_t start_ts = get_us();

		h_data_t hd{ 0 };
		if (s -> get_frame(E_RGB, 0, &hd.ts, &hd.width, &hd.height, &hd.data, &hd.len, true)) {
			if (!prev_frame)
				prev_frame = (uint8_t *)allocate_aligned(hd.len);

			if (filters) {
				apply_filters(nullptr, this, filters, prev_frame, hd.data, hd.ts, hd.width, hd.height);
				memcpy(prev_frame, hd.data, hd.len);
			}

			frames_lock.lock();

			while(frames.size() >= n_frames) {
				free(frames.front().data);
				frames.erase(frames.begin());
			}

			frames.push_back(hd);

			frames_lock.unlock();
		}

		if (interval) {
			uint64_t end_ts = get_us();
			int64_t left = interval - (end_ts - start_ts);

			if (left > 0)
				myusleep(left);
		}
	}

	s -> unregister_user();

	log(id, LL_INFO, "source_delay thread terminating");
}

bool source_delay::get_frame(const encoding_t pe, const int jpeg_quality, uint64_t *ts, int *width, int *height, uint8_t **frame, size_t *frame_len, const bool repeat)
{
	h_data_t hdchosen{ 0 };

	frames_lock.lock_shared();

	if (!frames.empty())
		hdchosen = frames.at(0);

	if (hdchosen.data == nullptr) {
		frames_lock.unlock_shared();

		*ts = 0;
		*frame = nullptr;
		*frame_len = 0;

		do_exit_failure();

		return false;
	}

	*width = hdchosen.width;
	*height = hdchosen.height;
	this->ts = *ts = hdchosen.ts;

	if (pe == E_JPEG) {
		bool rc = write_JPEG_memory(get_meta(), *width, *height, jpeg_quality, hdchosen.data, (char **)frame, frame_len);
		frames_lock.unlock_shared();

		if (!rc)
			do_exit_failure();

		return rc;
	}

	if (pe == E_RGB) {
		*frame = (uint8_t *)allocate_aligned(hdchosen.len);
		memcpy(*frame, hdchosen.data, hdchosen.len);
		*frame_len = hdchosen.len;
		frames_lock.unlock_shared();
		return true;
	}

	frames_lock.unlock_shared();

	do_exit_failure();

	return false;
}

uint64_t source_delay::get_current_ts()
{
	uint64_t rc = 0;

	cpt(pthread_mutex_lock(&lock));

	if (!frames.empty())
		rc = ts = frames.at(0).ts;

	cpt(pthread_mutex_unlock(&lock));

	return rc;
}
